import asyncio
from pathlib import Path

import pytest

from swerex.exceptions import (
    BashIncorrectSyntaxError,
    CommandTimeoutError,
    NonZeroExitCodeError,
    SessionDoesNotExistError,
)
from swerex.runtime.abstract import (
    CloseBashSessionRequest,
    CreateBashSessionRequest,
    ReadFileRequest,
    UploadRequest,
    WriteFileRequest,
)
from swerex.runtime.remote import RemoteRuntime

from .conftest import _Action as A
from .conftest import _Command as C


async def test_server_alive(remote_runtime: RemoteRuntime):
    assert await remote_runtime.is_alive()


async def test_server_dead():
    r = RemoteRuntime(host="http://doesnotexistadsfasdfasdf234123qw34.com", auth_token="")
    assert not await r.is_alive()


async def test_read_write_file(remote_runtime: RemoteRuntime, tmp_path: Path):
    path = tmp_path / "test.txt"
    await remote_runtime.write_file(WriteFileRequest(path=str(path), content="test"))
    assert path.read_text() == "test"
    assert (await remote_runtime.read_file(ReadFileRequest(path=str(path)))).content == "test"


async def test_read_non_existent_file(remote_runtime: RemoteRuntime):
    with pytest.raises(FileNotFoundError):
        await remote_runtime.read_file(ReadFileRequest(path="non_existent.txt"))


async def test_execute_command(remote_runtime: RemoteRuntime):
    assert (await remote_runtime.execute(C(command="echo 'hello world'", shell=True))).stdout == "hello world\n"


async def test_execute_command_shell_false(remote_runtime: RemoteRuntime):
    assert (await remote_runtime.execute(C(command=["echo", "hello world"], shell=False))).stdout == "hello world\n"


async def test_execute_command_timeout(remote_runtime: RemoteRuntime):
    with pytest.raises(CommandTimeoutError):
        await remote_runtime.execute(C(command=["sleep", "10"], timeout=0.1))


async def test_create_close_shell(remote_runtime: RemoteRuntime):
    await remote_runtime.create_session(CreateBashSessionRequest())
    await remote_runtime.close_session(CloseBashSessionRequest())


async def test_run_in_shell(runtime_with_default_session: RemoteRuntime):
    r = await runtime_with_default_session.run_in_session(A(command="echo 'hello world'", check=True))
    r = await runtime_with_default_session.run_in_session(A(command="doesntexit"))
    assert r.exit_code == 127
    r = await runtime_with_default_session.run_in_session(A(command="false && true"))
    assert r.exit_code == 1
    r = await runtime_with_default_session.run_in_session(A(command="false || true", check=True))


async def test_run_in_shell_check_exit_code(runtime_with_default_session: RemoteRuntime):
    with pytest.raises(NonZeroExitCodeError):
        await runtime_with_default_session.run_in_session(A(command="false", check=True))


async def test_run_in_shell_non_existent_session(remote_runtime: RemoteRuntime):
    with pytest.raises(SessionDoesNotExistError):
        await remote_runtime.run_in_session(A(command="echo 'hello world'", session="non_existent"))


async def test_close_shell_non_existent_session(remote_runtime: RemoteRuntime):
    with pytest.raises(SessionDoesNotExistError):
        await remote_runtime.close_session(CloseBashSessionRequest(session="non_existent"))


async def test_close_shell_twice(remote_runtime: RemoteRuntime):
    n = "mynewsession"
    await remote_runtime.create_session(CreateBashSessionRequest(session=n))
    await remote_runtime.close_session(CloseBashSessionRequest(session=n))
    with pytest.raises(SessionDoesNotExistError):
        await remote_runtime.close_session(CloseBashSessionRequest(session=n))


async def test_run_in_shell_timeout(runtime_with_default_session: RemoteRuntime):
    with pytest.raises(CommandTimeoutError):
        await runtime_with_default_session.run_in_session(A(command="sleep 10", timeout=0.1))


async def test_run_in_shell_interactive_command(runtime_with_default_session: RemoteRuntime):
    await runtime_with_default_session.run_in_session(A(command="python", is_interactive_command=True, expect=[">>> "]))
    await runtime_with_default_session.run_in_session(
        A(command="print('hello world')", is_interactive_command=True, expect=[">>> "])
    )
    await runtime_with_default_session.run_in_session(A(command="quit()\n", is_interactive_quit=True, check=True))


async def test_run_in_shell_multiple_interactive_and_normal_commands(runtime_with_default_session: RemoteRuntime):
    run = runtime_with_default_session
    await run.run_in_session(A(command="ls", check=True))
    await run.run_in_session(A(command="python", is_interactive_command=True, expect=[">>> "]))

    r = await run.run_in_session(A(command="print('hello world')", is_interactive_command=True, expect=[">>> "]))
    assert "hello world" in r.output

    await run.run_in_session(A(command="quit()\n", is_interactive_quit=True, check=True))

    r = await run.run_in_session(A(command="echo 'hello world'", check=True))
    assert "hello world" in r.output

    await run.run_in_session(A(command="python", is_interactive_command=True, expect=[">>> "]))
    await run.run_in_session(A(command="print('hello world')", is_interactive_command=True, expect=[">>> "]))
    await run.run_in_session(A(command="quit()\n", is_interactive_quit=True, check=True))
    r = await run.run_in_session(A(command="echo 'hello world'", check=True))
    assert "hello world" in r.output


async def test_run_in_shell_interactive_command_timeout(runtime_with_default_session: RemoteRuntime):
    with pytest.raises(CommandTimeoutError):
        await runtime_with_default_session.run_in_session(
            A(command="python", is_interactive_command=True, expect=["WONTHITTHIS"], timeout=0.1)
        )


async def test_write_to_non_existent_directory(remote_runtime: RemoteRuntime, tmp_path: Path):
    non_existent_dir = tmp_path / "non_existent_dir" / "test.txt"
    await remote_runtime.write_file(WriteFileRequest(path=str(non_existent_dir), content="test"))


async def test_read_large_file(remote_runtime: RemoteRuntime, tmp_path: Path):
    large_file = tmp_path / "large_file.txt"
    content = "x" * 1024 * 1024  # 1 MB of data
    large_file.write_text(content)

    response = await remote_runtime.read_file(ReadFileRequest(path=str(large_file)))
    assert len(response.content) == len(content)


async def test_multiple_isolated_shells(remote_runtime: RemoteRuntime):
    await remote_runtime.create_session(CreateBashSessionRequest(session="shell1"))
    await remote_runtime.create_session(CreateBashSessionRequest(session="shell2"))

    await asyncio.gather(
        remote_runtime.run_in_session(A(command="x=42", session="shell1", check=True)),
        remote_runtime.run_in_session(A(command="y=24", session="shell2", check=True)),
    )

    response1 = await remote_runtime.run_in_session(A(command="echo $x", session="shell1", check=True))
    response2 = await remote_runtime.run_in_session(A(command="echo $y", session="shell2", check=True))

    assert response1.output.strip() == "42"
    assert response2.output.strip() == "24"

    response3 = await remote_runtime.run_in_session(A(command="echo $y", session="shell1", check=True))
    response4 = await remote_runtime.run_in_session(A(command="echo $x", session="shell2", check=True))

    assert response3.output.strip() == ""
    assert response4.output.strip() == ""

    await remote_runtime.close_session(CloseBashSessionRequest(session="shell1"))
    await remote_runtime.close_session(CloseBashSessionRequest(session="shell2"))


async def test_empty_command(remote_runtime: RemoteRuntime):
    await remote_runtime.execute(C(command="", shell=True, check=True))
    await remote_runtime.execute(C(command="\n", shell=True, check=True))


async def test_command_fails_check_exit_code(runtime_with_default_session: RemoteRuntime):
    with pytest.raises(NonZeroExitCodeError):
        await runtime_with_default_session.run_in_session(A(command="false", check=True))


async def test_empty_command_in_shell(runtime_with_default_session: RemoteRuntime):
    await runtime_with_default_session.run_in_session(A(command="", check=True))
    await runtime_with_default_session.run_in_session(A(command="\n", check=True))
    await runtime_with_default_session.run_in_session(A(command="\n\n \n", check=True))


async def test_command_with_linebreaks(runtime_with_default_session: RemoteRuntime):
    await runtime_with_default_session.run_in_session(A(command="\n echo 'test'\n\n", check=True))


async def test_multiple_commands_with_linebreaks_in_shell(runtime_with_default_session: RemoteRuntime):
    r = await runtime_with_default_session.run_in_session(
        A(command="\n\n\n echo 'test1' \n  \n \n echo 'test2' \n\n\n", check=True)
    )
    assert r.output.splitlines() == ["test1", "test2"]


async def test_bash_multiline_command_eof(runtime_with_default_session: RemoteRuntime):
    command = "\n".join(["python <<EOF", "print('hello world')", "print('hello world 2')", "EOF"])
    r = await runtime_with_default_session.run_in_session(A(command=command, check=True))
    assert "hello world" in r.output
    assert "hello world 2" in r.output


async def test_run_in_shell_subshell_command(runtime_with_default_session: RemoteRuntime):
    await runtime_with_default_session.run_in_session(A(command="(sleep 10) &", check=True))


async def test_run_just_comment(runtime_with_default_session: RemoteRuntime):
    r = await runtime_with_default_session.run_in_session(A(command="# echo 'hello world'", check=True))
    assert r.output == ""


async def test_run_in_shell_multiple_commands(runtime_with_default_session: RemoteRuntime):
    r = await runtime_with_default_session.run_in_session(
        A(command="echo 'hello world'; echo 'hello again'", check=True)
    )
    assert r.output.splitlines() == ["hello world", "hello again"]
    r = await runtime_with_default_session.run_in_session(
        A(command="echo 'hello world' && echo 'hello again'", check=True)
    )
    assert r.output.splitlines() == ["hello world", "hello again"]


async def test_run_in_shell_while_loop(runtime_with_default_session: RemoteRuntime):
    r = await runtime_with_default_session.run_in_session(
        A(command="for i in {1..3};\n do echo 'hello world';\n done", check=True)
    )
    assert r.output.splitlines() == ["hello world"] * 3


async def test_run_in_shell_bashlex_errors(runtime_with_default_session: RemoteRuntime):
    # One of the bugs in bashlex
    await runtime_with_default_session.run_in_session(A(command="[[ $env == $env ]]", check=True))


async def test_run_shell_check_exit_code(runtime_with_default_session: RemoteRuntime):
    r = await runtime_with_default_session.run_in_session(A(command="/bin/bash -n <<'EOF'\necho 'hello world'\nEOF"))
    assert r.exit_code == 0


async def test_with_bashlex_errors(runtime_with_default_session: RemoteRuntime):
    r = await runtime_with_default_session.run_in_session(A(command="echo 'hw';A=();echo 'asdf'", check=True))
    assert "hw" in r.output
    assert "asdf" in r.output


async def test_upload_file(runtime_with_default_session: RemoteRuntime, tmp_path: Path):
    file_path = tmp_path / "source.txt"
    file_path.write_text("test")
    tmp_target = tmp_path / "target.txt"
    await runtime_with_default_session.upload(UploadRequest(source_path=str(file_path), target_path=str(tmp_target)))
    result = await runtime_with_default_session.read_file(ReadFileRequest(path=str(tmp_target)))
    assert result.content == "test"


async def test_upload_directory(runtime_with_default_session: RemoteRuntime, tmp_path: Path):
    dir_path = tmp_path / "source_dir"
    dir_path.mkdir()
    (dir_path / "file1.txt").write_text("test1")
    (dir_path / "file2.txt").write_text("test2")
    tmp_target = tmp_path / "target_dir"
    await runtime_with_default_session.upload(UploadRequest(source_path=str(dir_path), target_path=str(tmp_target)))

    assert (
        await runtime_with_default_session.read_file(ReadFileRequest(path=str(tmp_target / "file1.txt")))
    ).content == "test1"
    assert (
        await runtime_with_default_session.read_file(ReadFileRequest(path=str(tmp_target / "file2.txt")))
    ).content == "test2"


async def test_fail_bashlex_errors(runtime_with_default_session: RemoteRuntime):
    r = await runtime_with_default_session.run_in_session(A(command="A=(); false"))
    assert r.exit_code == 1
    assert r.output == ""


async def test_check_bash_command_invalid(runtime_with_default_session: RemoteRuntime):
    with pytest.raises(BashIncorrectSyntaxError):
        await runtime_with_default_session.run_in_session(A(command="(a"))



import asyncio
import os
import time
import uuid
from pathlib import Path, PurePath
from typing import Any, Self

import boto3
import modal
from botocore.exceptions import NoCredentialsError

from swerex import PACKAGE_NAME, REMOTE_EXECUTABLE_NAME
from swerex.deployment.abstract import AbstractDeployment
from swerex.deployment.config import ModalDeploymentConfig
from swerex.exceptions import DeploymentNotStartedError
from swerex.runtime.abstract import IsAliveResponse
from swerex.runtime.config import RemoteRuntimeConfig
from swerex.runtime.remote import RemoteRuntime
from swerex.utils.log import get_logger
from swerex.utils.wait import _wait_until_alive

__all__ = ["ModalDeployment"]


def _get_modal_user() -> str:
    # not sure how to get the user from the modal api
    return modal.config._profile  # type: ignore


class _ImageBuilder:
    """_ImageBuilder.auto() is used by ModalDeployment"""

    def __init__(self):
        self.logger = get_logger("_image_builder")

    def from_file(self, image: PurePath, *, build_context: PurePath | None = None) -> modal.Image:
        self.logger.info(f"Building image from file {image}")
        if build_context is None:
            build_context = Path(image).resolve().parent
        build_context = Path(build_context)
        self.logger.debug(f"Using build context {build_context}")
        context_mount = modal.Mount.from_local_dir(
            local_path=build_context,
            remote_path=".",  # to current WORKDIR
        )
        return modal.Image.from_dockerfile(str(image), context_mount=context_mount)

    def from_registry(self, image: str) -> modal.Image:
        self.logger.info(f"Building image from docker registry {image}")
        if os.environ.get("DOCKER_USERNAME") and os.environ.get("DOCKER_PASSWORD"):
            secret = modal.Secret.from_dict(
                {
                    "DOCKER_USERNAME": os.environ["DOCKER_USERNAME"],
                    "DOCKER_PASSWORD": os.environ["DOCKER_PASSWORD"],
                }
            )
            secrets = [secret]
            self.logger.debug("Docker login credentials were provided")
        else:
            self.logger.warning("DOCKER_USERNAME and DOCKER_PASSWORD not set. Using public images.")
            secrets = None
        return modal.Image.from_registry(image, secrets=secrets)

    def from_ecr(self, image: str) -> modal.Image:
        self.logger.info(f"Building image from ECR {image}")
        try:
            session = boto3.Session()
            credentials = session.get_credentials()
            aws_access_key_id = credentials.access_key
            aws_secret_access_key = credentials.secret_key
            secret = modal.Secret.from_dict(
                {
                    "AWS_ACCESS_KEY_ID": aws_access_key_id,
                    "AWS_SECRET_ACCESS_KEY": aws_secret_access_key,
                }
            )
            return modal.Image.from_ecr(  # type: ignore
                image,
                secrets=[secret],
            )
        except NoCredentialsError as e:
            msg = "AWS credentials not found. Please configure your AWS credentials."
            raise ValueError(msg) from e

    def ensure_pipx_installed(self, image: modal.Image) -> modal.Image:
        image = image.apt_install("pipx")
        return image.run_commands("pipx ensurepath")

    def auto(self, image_spec: str | modal.Image | PurePath) -> modal.Image:
        if isinstance(image_spec, modal.Image):
            image = image_spec
        elif isinstance(image_spec, PurePath) and not Path(image_spec).is_file():
            msg = f"File {image_spec} does not exist"
            raise FileNotFoundError(msg)
        elif Path(image_spec).is_file():
            image = self.from_file(Path(image_spec))
        elif "amazonaws.com" in image_spec:  # type: ignore
            image = self.from_ecr(image_spec)  # type: ignore
        else:
            image = self.from_registry(image_spec)  # type: ignore

        return self.ensure_pipx_installed(image)


class ModalDeployment(AbstractDeployment):
    # Leave the constructor args for now, because image can take a modal.Image
    # but we don't want to make this part of the config class because it would
    # force us to have modal installed/import it...
    def __init__(
        self,
        image: str | modal.Image | PurePath,
        startup_timeout: float = 0.4,
        runtime_timeout: float = 1800.0,
        modal_sandbox_kwargs: dict[str, Any] | None = None,
    ):
        """Deployment for modal.com. The deployment will only start when the
        `start` method is being called.

        Args:
            image: Image to use for the deployment. One of the following:
                1. `modal.Image` object
                2. Path to a Dockerfile
                3. Dockerhub image name (e.g. `python:3.11-slim`)
                4. ECR image name (e.g. `123456789012.dkr.ecr.us-east-1.amazonaws.com/my-image:tag`)
            startup_timeout: The time to wait for the runtime to start.
            runtime_timeout: The runtime timeout.
            modal_sandbox_kwargs: Additional arguments to pass to `modal.Sandbox.create`
        """
        self._image = _ImageBuilder().auto(image)
        self._runtime: RemoteRuntime | None = None
        self._startup_timeout = startup_timeout
        self._sandbox: modal.Sandbox | None = None
        self._port = 8880
        self.logger = get_logger("deploy")
        self._app = modal.App.lookup("swe-rex", create_if_missing=True)
        self._user = _get_modal_user()
        self._runtime_timeout = runtime_timeout
        if modal_sandbox_kwargs is None:
            modal_sandbox_kwargs = {}
        self._modal_kwargs = modal_sandbox_kwargs

    @classmethod
    def from_config(cls, config: ModalDeploymentConfig) -> Self:
        return cls(
            image=config.image,
            startup_timeout=config.startup_timeout,
            runtime_timeout=config.runtime_timeout,
            modal_sandbox_kwargs=config.modal_sandbox_kwargs,
        )

    def _get_token(self) -> str:
        return str(uuid.uuid4())

    async def is_alive(self, *, timeout: float | None = None) -> IsAliveResponse:
        """Checks if the runtime is alive. The return value can be
        tested with bool().

        Raises:
            DeploymentNotStartedError: If the deployment was not started.
        """
        if self._runtime is None or self._sandbox is None:
            raise DeploymentNotStartedError()
        if self._sandbox.poll() is not None:
            msg = "Container process terminated."
            output = "stdout:\n" + self._sandbox.stdout.read()  # type: ignore
            output += "\nstderr:\n" + self._sandbox.stderr.read()  # type: ignore
            msg += "\n" + output
            raise RuntimeError(msg)
        return await self._runtime.is_alive(timeout=timeout)

    async def _wait_until_alive(self, timeout: float = 10.0):
        assert self._runtime is not None
        return await _wait_until_alive(self.is_alive, timeout=timeout, function_timeout=self._runtime._config.timeout)

    def _start_swerex_cmd(self, token: str) -> str:
        """Start swerex-server on the remote. If swerex is not installed arelady,
        install pipx and then run swerex-server with pipx run
        """
        rex_args = f"--port {self._port} --auth-token {token}"
        return f"{REMOTE_EXECUTABLE_NAME} {rex_args} || pipx run {PACKAGE_NAME} {rex_args}"

    def get_modal_log_url(self) -> str:
        """Returns URL to modal logs

        Raises:
            DeploymentNotStartedError: If the deployment was not started.
        """
        return f"https://modal.com/apps/{self._user}/main/deployed/{self.app.name}?activeTab=logs&taskId={self.sandbox._get_task_id()}"

    async def start(
        self,
    ):
        """Starts the runtime."""
        self.logger.info("Starting modal sandbox")
        t0 = time.time()
        token = self._get_token()
        self._sandbox = modal.Sandbox.create(
            "/bin/bash",
            "-c",
            self._start_swerex_cmd(token),
            image=self._image,
            timeout=int(self._runtime_timeout),
            unencrypted_ports=[self._port],
            app=self._app,
            **self._modal_kwargs,
        )
        tunnel = self._sandbox.tunnels()[self._port]
        elapsed_sandbox_creation = time.time() - t0
        self.logger.info(f"Sandbox ({self._sandbox.object_id}) created in {elapsed_sandbox_creation:.2f}s")
        self.logger.info(f"Check sandbox logs at {self.get_modal_log_url()}")
        self.logger.info(f"Sandbox created with id {self._sandbox.object_id}")
        await asyncio.sleep(1)
        self.logger.info(f"Starting runtime at {tunnel.url}")
        self._runtime = RemoteRuntime.from_config(
            RemoteRuntimeConfig(host=tunnel.url, timeout=self._runtime_timeout, auth_token=token)
        )
        remaining_startup_timeout = max(0, self._startup_timeout - elapsed_sandbox_creation)
        t1 = time.time()
        await self._wait_until_alive(timeout=remaining_startup_timeout)
        self.logger.info(f"Runtime started in {time.time() - t1:.2f}s")

    async def stop(self):
        """Stops the runtime."""
        if self._runtime is not None:
            await self._runtime.close()
            self._runtime = None
        if self._sandbox is not None and not self._sandbox.poll():
            self._sandbox.terminate()
        self._sandbox = None
        self._app = None

    @property
    def runtime(self) -> RemoteRuntime:
        """Returns the runtime if running.

        Raises:
            DeploymentNotStartedError: If the deployment was not started.
        """
        if self._runtime is None:
            raise DeploymentNotStartedError()
        return self._runtime

    @property
    def app(self) -> modal.App:
        """Returns the modal app

        Raises:
            DeploymentNotStartedError: If the deployment is not started.
        """
        if self._app is None:
            raise DeploymentNotStartedError()
        return self._app

    @property
    def sandbox(self) -> modal.Sandbox:
        """Returns the modal sandbox

        Raises:
            DeploymentNotStartedError: If the deployment is not started.
        """
        if self._sandbox is None:
            raise DeploymentNotStartedError()
        return self._sandbox



import pytest

from swerex.deployment.docker import DockerDeployment
from swerex.utils.free_port import find_free_port


async def test_docker_deployment():
    port = find_free_port()
    print(f"Using port {port} for the docker deployment")
    d = DockerDeployment(image="swe-rex-test:latest", port=port)
    with pytest.raises(RuntimeError):
        await d.is_alive()
    await d.start()
    assert await d.is_alive()
    await d.stop()



import pytest

from swerex.exceptions import BashIncorrectSyntaxError
from swerex.runtime.local import _check_bash_command, _split_bash_command


def test_split_bash_command_normal():
    assert _split_bash_command("cmd1\ncmd2") == ["cmd1", "cmd2"]


def test_split_bash_command_escaped_newline():
    assert _split_bash_command("cmd1\\\n asdf") == ["cmd1\\\n asdf"]


def test_split_bash_command_heredoc():
    assert _split_bash_command("cmd1<<EOF\na\nb\nEOF") == ["cmd1<<EOF\na\nb\nEOF"]
    assert _split_bash_command("cmd1<<EOF\na\nb\nEOF\ncmd2<<EOF\nd\ne\nEOF") == [
        "cmd1<<EOF\na\nb\nEOF",
        "cmd2<<EOF\nd\ne\nEOF",
    ]


def test_split_bash_command_multiple_commands():
    assert _split_bash_command("cmd1\ncmd2\ncmd3") == ["cmd1", "cmd2", "cmd3"]


def test_split_bash_command_multiple_commands_with_linebreaks():
    assert _split_bash_command("cmd1\n\ncmd2\n\ncmd3") == ["cmd1", "cmd2", "cmd3"]


def test_split_bash_command_multiple_commands_with_heredocs():
    assert _split_bash_command("cmd1<<EOF\na\nb\nEOF\ncmd2<<EOF\nd\ne\nEOF") == [
        "cmd1<<EOF\na\nb\nEOF",
        "cmd2<<EOF\nd\ne\nEOF",
    ]


def test_split_bash_command_multilline_blank_line_heredoc():
    assert _split_bash_command("cmd1<<EOF\na\nb\n\n\nEOF\ncmd2<<EOF\nd\ne\nEOF") == [
        "cmd1<<EOF\na\nb\n\n\nEOF",
        "cmd2<<EOF\nd\ne\nEOF",
    ]


def test_split_bash_command_all_blank_lines():
    assert _split_bash_command("\n\n\n") == []


def test_split_bash_command_quotation_marks():
    assert _split_bash_command('cmd1 "a\nb"') == [
        'cmd1 "a\nb"',
    ]
    assert _split_bash_command("cmd1 'a\nb'") == [
        "cmd1 'a\nb'",
    ]


def test_split_command_with_heredoc_quotations():
    assert _split_bash_command('cmd1 <<EOF\n"a\nb"\nEOF') == [
        'cmd1 <<EOF\n"a\nb"\nEOF',
    ]
    assert _split_bash_command("cmd1 <<EOF\n'a\nb'\nEOF") == [
        "cmd1 <<EOF\n'a\nb'\nEOF",
    ]


def test_check_bash_command_invalid():
    with pytest.raises(BashIncorrectSyntaxError):
        _check_bash_command("(a")
    with pytest.raises(BashIncorrectSyntaxError):
        _check_bash_command("a='")
    with pytest.raises(BashIncorrectSyntaxError):
        _check_bash_command("for")


def test_check_bash_command_valid():
    _check_bash_command("(a)")
    _check_bash_command("a=''")



import requests

from tests.conftest import RemoteServer

headers = {"X-API-Key": "your_secret_api_key_here"}


def test_is_alive(remote_server: RemoteServer):
    response = requests.get(f"http://127.0.0.1:{remote_server.port}/is_alive", headers=remote_server.headers)
    print(response.json())
    assert response.json()["is_alive"]


def test_hello_world(remote_server: RemoteServer):
    assert (
        requests.get(f"http://127.0.0.1:{remote_server.port}/", headers=remote_server.headers).json()["message"]
        == "hello world"
    )


def test_unauthenticated_request(remote_server: RemoteServer):
    for endpoint in [
        "/is_alive",
        "/",
        "/create_session",
        "/run_in_session",
        "/close_session",
        "/execute",
        "/read_file",
        "/write_file",
        "/upload",
        "/close",
    ]:
        response = requests.get(f"http://127.0.0.1:{remote_server.port}/{endpoint}")
        assert response.status_code == 403



from typing import Any, Self

from swerex.runtime.abstract import (
    AbstractRuntime,
    Action,
    BashObservation,
    CloseBashSessionResponse,
    CloseResponse,
    CloseSessionRequest,
    CloseSessionResponse,
    Command,
    CommandResponse,
    CreateBashSessionResponse,
    CreateSessionRequest,
    CreateSessionResponse,
    IsAliveResponse,
    Observation,
    ReadFileRequest,
    ReadFileResponse,
    UploadRequest,
    UploadResponse,
    WriteFileRequest,
    WriteFileResponse,
)
from swerex.runtime.config import DummyRuntimeConfig


class DummyRuntime(AbstractRuntime):
    def __init__(
        self,
        **kwargs: Any,
    ):
        """This runtime does nothing.
        Useful for testing.

        Args:
            **kwargs: Keyword arguments (see `DummyRuntimeConfig` for details).
        """
        self.run_in_session_outputs: list[BashObservation] | BashObservation = BashObservation(exit_code=0)
        """Predefine returns of run_in_session. If set to list, will pop from list, else will 
        return the same value.
        """
        self._config = DummyRuntimeConfig(**kwargs)

    @classmethod
    def from_config(cls, config: DummyRuntimeConfig) -> Self:
        return cls(**config.model_dump())

    async def is_alive(self, *, timeout: float | None = None) -> IsAliveResponse:
        return IsAliveResponse(is_alive=True)

    async def create_session(self, request: CreateSessionRequest) -> CreateSessionResponse:
        if request.session_type == "bash":
            return CreateBashSessionResponse()
        msg = f"Unknown session type: {request.session_type}"
        raise ValueError(msg)

    async def run_in_session(self, action: Action) -> Observation:
        if isinstance(self.run_in_session_outputs, list):
            return self.run_in_session_outputs.pop(0)
        return self.run_in_session_outputs

    async def close_session(self, request: CloseSessionRequest) -> CloseSessionResponse:
        if request.session_type == "bash":
            return CloseBashSessionResponse()
        msg = f"Unknown session type: {request.session_type}"
        raise ValueError(msg)

    async def execute(self, command: Command) -> CommandResponse:
        return CommandResponse(exit_code=0)

    async def read_file(self, request: ReadFileRequest) -> ReadFileResponse:
        return ReadFileResponse()

    async def write_file(self, request: WriteFileRequest) -> WriteFileResponse:
        return WriteFileResponse()

    async def upload(self, request: UploadRequest) -> UploadResponse:
        return UploadResponse()

    async def close(self) -> CloseResponse:
        return CloseResponse()



from pathlib import Path

import pytest

from swerex.deployment.modal import ModalDeployment, _ImageBuilder


@pytest.mark.slow
async def test_modal_deployment_from_docker_with_swerex_installed():
    dockerfile = Path(__file__).parent / "swe_rex_test.Dockerfile"
    image = _ImageBuilder().from_file(dockerfile, build_context=Path(__file__).resolve().parent.parent)
    d = ModalDeployment(image=image, startup_timeout=60)
    with pytest.raises(RuntimeError):
        await d.is_alive()
    await d.start()
    assert await d.is_alive()
    await d.stop()


@pytest.mark.slow
async def test_modal_deployment_from_docker_string():
    d = ModalDeployment(image="python:3.11-slim", startup_timeout=60)
    await d.start()
    assert await d.is_alive()
    await d.stop()



from pydantic import BaseModel

from swerex.runtime.abstract import AbstractRuntime


class LocalRuntimeConfig(BaseModel):
    """Configuration for a local runtime."""


class RemoteRuntimeConfig(BaseModel):
    auth_token: str
    """The token to use for authentication."""
    host: str = "http://127.0.0.1"
    """The host to connect to."""
    port: int | None = None
    """The port to connect to."""
    timeout: float = 0.15


class DummyRuntimeConfig(BaseModel):
    """Configuration for a dummy runtime."""


RuntimeConfig = LocalRuntimeConfig | RemoteRuntimeConfig | DummyRuntimeConfig


def get_runtime(config: RuntimeConfig) -> AbstractRuntime:
    if isinstance(config, LocalRuntimeConfig):
        from swerex.runtime.local import LocalRuntime

        return LocalRuntime.from_config(config)
    if isinstance(config, RemoteRuntimeConfig):
        from swerex.runtime.remote import RemoteRuntime

        return RemoteRuntime.from_config(config)
    if isinstance(config, DummyRuntimeConfig):
        from swerex.runtime.dummy import DummyRuntime

        return DummyRuntime.from_config(config)

    msg = f"Unknown runtime type: {type(config)}"
    raise ValueError(msg)



import time
from collections.abc import Callable


async def _wait_until_alive(
    function: Callable, timeout: float = 10.0, function_timeout: float | None = 0.1, sleep: float = 0.25
):
    """Wait until the function returns a truthy value.

    Args:
        function: The function to wait for.
        timeout: The maximum time to wait.
        function_timeout: The timeout passed to the function.
        sleep: The time to sleep between attempts.

    Raises:
        TimeoutError
    """
    end_time = time.time() + timeout
    n_attempts = 0
    await_response = None
    while time.time() < end_time:
        await_response = await function(timeout=function_timeout)
        if await_response:
            return
        time.sleep(sleep)
        n_attempts += 1
    last_response_message = await_response.message if await_response else None
    msg = (
        f"Runtime did not start within {timeout}s (tried to connect {n_attempts} times). "
        f"The last await response was:\n{last_response_message}"
    )
    raise TimeoutError(msg)



#!/usr/bin/env python3

import argparse
import shutil
import tempfile
import traceback
import zipfile
from pathlib import Path

from fastapi import FastAPI, File, Form, HTTPException, Request, UploadFile
from fastapi.exception_handlers import http_exception_handler
from fastapi.responses import JSONResponse
from fastapi.security import APIKeyHeader
from starlette.exceptions import HTTPException as StarletteHTTPException

from swerex.runtime.abstract import (
    BashAction,
    CloseResponse,
    CloseSessionRequest,
    Command,
    CreateSessionRequest,
    ReadFileRequest,
    UploadResponse,
    WriteFileRequest,
    _ExceptionTransfer,
)
from swerex.runtime.local import LocalRuntime

app = FastAPI()
runtime = LocalRuntime()

AUTH_TOKEN = ""
api_key_header = APIKeyHeader(name="X-API-Key")


def serialize_model(model):
    return model.model_dump() if hasattr(model, "model_dump") else model.dict()


@app.middleware("http")
async def authenticate(request: Request, call_next):
    """Authenticate requests with an API key (if set)."""
    if AUTH_TOKEN:
        api_key = await api_key_header(request)
        if api_key != AUTH_TOKEN:
            raise HTTPException(status_code=401, detail="Invalid API Key")
    return await call_next(request)


@app.exception_handler(Exception)
async def exception_handler(request: Request, exc: Exception):
    """We catch exceptions that are thrown by the runtime, serialize them to JSON and
    return them to the client so they can reraise them in their own code.
    """
    if isinstance(exc, (HTTPException, StarletteHTTPException)):
        return await http_exception_handler(request, exc)
    _exc = _ExceptionTransfer(
        message=str(exc), class_path=type(exc).__module__ + "." + type(exc).__name__, traceback=traceback.format_exc()
    )
    return JSONResponse(status_code=511, content={"swerexception": _exc.model_dump()})


@app.get("/")
async def root():
    return {"message": "hello world"}


@app.get("/is_alive")
async def is_alive():
    return serialize_model(await runtime.is_alive())


@app.post("/create_session")
async def create_session(request: CreateSessionRequest):
    return serialize_model(await runtime.create_session(request))


@app.post("/run_in_session")
async def run(action: BashAction):
    return serialize_model(await runtime.run_in_session(action))


@app.post("/close_session")
async def close_session(request: CloseSessionRequest):
    return serialize_model(await runtime.close_session(request))


@app.post("/execute")
async def execute(command: Command):
    return serialize_model(await runtime.execute(command))


@app.post("/read_file")
async def read_file(request: ReadFileRequest):
    return serialize_model(await runtime.read_file(request))


@app.post("/write_file")
async def write_file(request: WriteFileRequest):
    return serialize_model(await runtime.write_file(request))


@app.post("/upload")
async def upload(
    file: UploadFile = File(...),
    target_path: str = Form(...),  # type: ignore
    unzip: bool = Form(False),
):
    target_path: Path = Path(target_path)
    target_path.parent.mkdir(parents=True, exist_ok=True)
    # First save the file to a temporary directory and potentially unzip it.
    with tempfile.TemporaryDirectory() as temp_dir:
        file_path = Path(temp_dir) / "temp_file_transfer"
        try:
            with open(file_path, "wb") as f:
                f.write(await file.read())
        finally:
            await file.close()
        if unzip:
            with zipfile.ZipFile(file_path, "r") as zip_ref:
                zip_ref.extractall(target_path)
            file_path.unlink()
        else:
            shutil.move(file_path, target_path)
    return UploadResponse()


@app.post("/close")
async def close():
    await runtime.close()
    return CloseResponse()


def main():
    import uvicorn

    parser = argparse.ArgumentParser(description="Run the SWE-ReX FastAPI server")
    parser.add_argument("--host", default="0.0.0.0", help="Host to bind the server to")
    parser.add_argument("--port", type=int, default=8000, help="Port to run the server on")
    parser.add_argument("--auth-token", default="", help="token to authenticate requests", required=True)

    args = parser.parse_args()
    global AUTH_TOKEN
    AUTH_TOKEN = args.auth_token
    uvicorn.run(app, host=args.host, port=args.port)


if __name__ == "__main__":
    main()



import pytest

from swerex.deployment.remote import RemoteDeployment
from swerex.exceptions import DeploymentNotStartedError
from tests.conftest import TEST_API_KEY


async def test_remote_deployment(remote_server):
    port = remote_server.port
    print(f"Using port {port} for the remote deployment")
    d = RemoteDeployment(port=port, auth_token=TEST_API_KEY)
    with pytest.raises(DeploymentNotStartedError):
        await d.is_alive()
    await d.start()
    assert await d.is_alive()
    await d.stop()



import shutil
import sys
import tempfile
import traceback
from pathlib import Path
from typing import Any, Self

import requests
from pydantic import BaseModel

from swerex.exceptions import SweRexception
from swerex.runtime.abstract import (
    AbstractRuntime,
    Action,
    CloseResponse,
    CloseSessionRequest,
    CloseSessionResponse,
    Command,
    CommandResponse,
    CreateSessionRequest,
    CreateSessionResponse,
    IsAliveResponse,
    Observation,
    ReadFileRequest,
    ReadFileResponse,
    UploadRequest,
    UploadResponse,
    WriteFileRequest,
    WriteFileResponse,
    _ExceptionTransfer,
)
from swerex.runtime.config import RemoteRuntimeConfig
from swerex.utils.log import get_logger
from swerex.utils.wait import _wait_until_alive

__all__ = ["RemoteRuntime", "RemoteRuntimeConfig"]


class RemoteRuntime(AbstractRuntime):
    def __init__(
        self,
        **kwargs: Any,
    ):
        """A runtime that connects to a remote server.

        Args:
            **kwargs: Keyword arguments to pass to the `RemoteRuntimeConfig` constructor.
        """
        self._config = RemoteRuntimeConfig(**kwargs)
        self.logger = get_logger("RR")
        if not self._config.host.startswith("http"):
            self.logger.warning("Host %s does not start with http, adding http://", self._config.host)
            self._config.host = f"http://{self._config.host}"

    @classmethod
    def from_config(cls, config: RemoteRuntimeConfig) -> Self:
        return cls(**config.model_dump())

    def _get_timeout(self, timeout: float | None = None) -> float:
        if timeout is None:
            return self._config.timeout
        return timeout

    @property
    def _headers(self) -> dict[str, str]:
        """Request headers to use for authentication."""
        if self._config.auth_token:
            return {"X-API-Key": self._config.auth_token}
        return {}

    @property
    def _api_url(self) -> str:
        if self._config.port is None:
            return self._config.host
        return f"{self._config.host}:{self._config.port}"

    def _handle_transfer_exception(self, exc_transfer: _ExceptionTransfer) -> None:
        """Reraise exceptions that were thrown on the remote."""
        if exc_transfer.traceback:
            print("Traceback: ", exc_transfer.traceback, file=sys.stderr)
        try:
            module, _, exc_name = exc_transfer.class_path.rpartition(".")
            exception = getattr(sys.modules[module], exc_name)
        except AttributeError:
            self.logger.error(f"Unknown exception class: {exc_transfer.class_path!r}")
            raise SweRexception(exc_transfer.message) from None
        raise exception(exc_transfer.message) from None

    def _handle_response_errors(self, response: requests.Response) -> None:
        """Raise exceptions found in the request response."""
        if response.status_code == 511:
            exc_transfer = _ExceptionTransfer(**response.json()["swerexception"])
            self._handle_transfer_exception(exc_transfer)
        response.raise_for_status()

    async def is_alive(self, *, timeout: float | None = None) -> IsAliveResponse:
        """Checks if the runtime is alive.

        Internal server errors are thrown, everything else just has us return False
        together with the message.
        """
        try:
            response = requests.get(
                f"{self._api_url}/is_alive", headers=self._headers, timeout=self._get_timeout(timeout)
            )
            if response.status_code == 200:
                return IsAliveResponse(**response.json())
            elif response.status_code == 511:
                exc_transfer = _ExceptionTransfer(**response.json()["swerexception"])
                self._handle_transfer_exception(exc_transfer)
            msg = (
                f"Status code {response.status_code} from {self._api_url}/is_alive. "
                f"Message: {response.json().get('detail')}"
            )
            return IsAliveResponse(is_alive=False, message=msg)
        except requests.RequestException:
            msg = f"Failed to connect to {self._config.host}\n"
            msg += traceback.format_exc()
            return IsAliveResponse(is_alive=False, message=msg)
        except Exception:
            msg = f"Failed to connect to {self._config.host}\n"
            msg += traceback.format_exc()
            return IsAliveResponse(is_alive=False, message=msg)

    async def wait_until_alive(self, *, timeout: float = 60.0):
        return await _wait_until_alive(self.is_alive, timeout=timeout)

    def _request(self, endpoint: str, request: BaseModel | None, output_class: Any):
        """Small helper to make requests to the server and handle errors and output."""
        response = requests.post(
            f"{self._api_url}/{endpoint}", json=request.model_dump() if request else None, headers=self._headers
        )
        self._handle_response_errors(response)
        return output_class(**response.json())

    async def create_session(self, request: CreateSessionRequest) -> CreateSessionResponse:
        """Creates a new session."""
        return self._request("create_session", request, CreateSessionResponse)

    async def run_in_session(self, action: Action) -> Observation:
        """Runs a command in a session."""
        return self._request("run_in_session", action, Observation)

    async def close_session(self, request: CloseSessionRequest) -> CloseSessionResponse:
        """Closes a shell session."""
        return self._request("close_session", request, CloseSessionResponse)

    async def execute(self, command: Command) -> CommandResponse:
        """Executes a command (independent of any shell session)."""
        return self._request("execute", command, CommandResponse)

    async def read_file(self, request: ReadFileRequest) -> ReadFileResponse:
        """Reads a file"""
        return self._request("read_file", request, ReadFileResponse)

    async def write_file(self, request: WriteFileRequest) -> WriteFileResponse:
        """Writes a file"""
        return self._request("write_file", request, WriteFileResponse)

    async def upload(self, request: UploadRequest) -> UploadResponse:
        """Uploads a file"""
        source = Path(request.source_path).resolve()
        self.logger.debug("Uploading file from %s to %s", request.source_path, request.target_path)
        if source.is_dir():
            with tempfile.TemporaryDirectory() as temp_dir:
                zip_path = Path(temp_dir) / "zipped_transfer.zip"
                shutil.make_archive(str(zip_path.with_suffix("")), "zip", source)
                self.logger.debug("Created zip file at %s", zip_path)
                files = {"file": zip_path.open("rb")}
                data = {"target_path": request.target_path, "unzip": "true"}
                response = requests.post(f"{self._api_url}/upload", files=files, data=data, headers=self._headers)
                self._handle_response_errors(response)
                return UploadResponse(**response.json())
        elif source.is_file():
            self.logger.debug("Uploading file from %s to %s", source, request.target_path)
            files = {"file": source.open("rb")}
            data = {"target_path": request.target_path, "unzip": "false"}
            response = requests.post(f"{self._api_url}/upload", files=files, data=data, headers=self._headers)
            self._handle_response_errors(response)
            return UploadResponse(**response.json())
        else:
            msg = f"Source path {source} is not a file or directory"
            raise ValueError(msg)

    async def close(self) -> CloseResponse:
        """Closes the runtime."""
        return self._request("close", None, CloseResponse)



class SweRexception(Exception):
    """Any exception that is raised by SWE-Rex."""


class SessionNotInitializedError(SweRexception, RuntimeError):
    """Raised if we try to run a command in a shell that is not initialized."""


class NonZeroExitCodeError(SweRexception, RuntimeError):
    """Can be raised if we execute a command in the shell and it has a non-zero exit code."""


class BashIncorrectSyntaxError(SweRexception, RuntimeError):
    """Before running a bash command, we check for syntax errors.
    This is the error message for those syntax errors.
    """


class CommandTimeoutError(SweRexception, RuntimeError, TimeoutError): ...


class NoExitCodeError(SweRexception, RuntimeError): ...


class SessionExistsError(SweRexception, ValueError): ...


class SessionDoesNotExistError(SweRexception, ValueError): ...


class DeploymentNotStartedError(SweRexception, RuntimeError):
    def __init__(self, message="Deployment not started"):
        super().__init__(message)



import shlex
import subprocess
import time
import uuid
from typing import Any, Self

from swerex import PACKAGE_NAME, REMOTE_EXECUTABLE_NAME
from swerex.deployment.abstract import AbstractDeployment
from swerex.deployment.config import DockerDeploymentConfig
from swerex.exceptions import DeploymentNotStartedError
from swerex.runtime.abstract import IsAliveResponse
from swerex.runtime.config import RemoteRuntimeConfig
from swerex.runtime.remote import RemoteRuntime
from swerex.utils.free_port import find_free_port
from swerex.utils.log import get_logger
from swerex.utils.wait import _wait_until_alive

__all__ = ["DockerDeployment", "DockerDeploymentConfig"]


def _is_image_available(image: str) -> bool:
    try:
        subprocess.check_call(["docker", "inspect", image], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        return True
    except subprocess.CalledProcessError:
        return False


def _pull_image(image: str) -> None:
    subprocess.check_call(["docker", "pull", image])


def _remove_image(image: str) -> None:
    subprocess.check_call(["docker", "rmi", image])


class DockerDeployment(AbstractDeployment):
    def __init__(
        self,
        **kwargs: Any,
    ):
        """Deployment to local docker image.

        Args:
            **kwargs: Keyword arguments (see `DockerDeploymentConfig` for details).
        """
        self._config = DockerDeploymentConfig(**kwargs)
        self._runtime: RemoteRuntime | None = None
        self._container_process = None
        self._container_name = None
        self.logger = get_logger("deploy")
        self._runtime_timeout = 0.15

    @classmethod
    def from_config(cls, config: DockerDeploymentConfig) -> Self:
        return cls(**config.model_dump())

    def _get_container_name(self) -> str:
        """Returns a unique container name based on the image name."""
        image_name_sanitized = "".join(c for c in self._config.image if c.isalnum() or c in "-_.")
        return f"{image_name_sanitized}-{uuid.uuid4()}"

    @property
    def container_name(self) -> str | None:
        return self._container_name

    async def is_alive(self, *, timeout: float | None = None) -> IsAliveResponse:
        """Checks if the runtime is alive. The return value can be
        tested with bool().

        Raises:
            DeploymentNotStartedError: If the deployment was not started.
        """
        if self._runtime is None:
            msg = "Runtime not started"
            raise RuntimeError(msg)
        if self._container_process is None:
            msg = "Container process not started"
            raise RuntimeError(msg)
        if self._container_process.poll() is not None:
            msg = "Container process terminated."
            output = "stdout:\n" + self._container_process.stdout.read().decode()  # type: ignore
            output += "\nstderr:\n" + self._container_process.stderr.read().decode()  # type: ignore
            msg += "\n" + output
            raise RuntimeError(msg)
        return await self._runtime.is_alive(timeout=timeout)

    async def _wait_until_alive(self, timeout: float = 10.0):
        try:
            return await _wait_until_alive(self.is_alive, timeout=timeout, function_timeout=self._runtime_timeout)
        except TimeoutError as e:
            self.logger.error("Runtime did not start within timeout. Here's the output from the container process.")
            assert self._container_process is not None
            self._container_process.terminate()
            self.logger.error(self._container_process.stdout.read().decode())  # type: ignore
            self.logger.error(self._container_process.stderr.read().decode())  # type: ignore
            raise e

    def _get_token(self) -> str:
        return str(uuid.uuid4())

    def _get_swerex_start_cmd(self, token: str) -> list[str]:
        rex_args = f"--auth-token {token}"
        pipx_install = "python3 -m pip install pipx && python3 -m pipx ensurepath"
        # todo: update
        # Need to wrap with /bin/sh -c to avoid having '&&' interpreted by the parent shell
        return [
            "/bin/sh",
            # "-l",
            "-c",
            f"{REMOTE_EXECUTABLE_NAME} {rex_args} || ({pipx_install} && pipx run {PACKAGE_NAME} {rex_args})",
        ]

    def _pull_image(self):
        if self._config.pull == "never":
            return
        if self._config.pull == "missing" and _is_image_available(self._config.image):
            return
        self.logger.info(f"Pulling image {self._config.image!r}")
        _pull_image(self._config.image)

    async def start(self):
        """Starts the runtime."""
        self._pull_image()
        if self._config.port is None:
            self._config.port = find_free_port()
        assert self._container_name is None
        self._container_name = self._get_container_name()
        token = self._get_token()
        cmds = [
            "docker",
            "run",
            "--rm",
            "-p",
            f"{self._config.port}:8000",
            *self._config.docker_args,
            "--name",
            self._container_name,
            self._config.image,
            *self._get_swerex_start_cmd(token),
        ]
        cmd_str = shlex.join(cmds)
        self.logger.info(
            f"Starting container {self._container_name} with image {self._config.image} serving on port {self._config.port}"
        )
        self.logger.debug(f"Command: {cmd_str!r}")
        # shell=True required for && etc.
        self._container_process = subprocess.Popen(cmds, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        self.logger.info(f"Starting runtime at {self._config.port}")
        self._runtime = RemoteRuntime.from_config(
            RemoteRuntimeConfig(port=self._config.port, timeout=self._runtime_timeout, auth_token=token)
        )
        t0 = time.time()
        await self._wait_until_alive(timeout=self._config.startup_timeout)
        self.logger.info(f"Runtime started in {time.time() - t0:.2f}s")

    async def stop(self):
        """Stops the runtime."""
        if self._runtime is not None:
            await self._runtime.close()
            self._runtime = None
        if self._container_process is not None:
            self._container_process.terminate()
            self._container_process = None
        self._container_name = None
        if self._config.remove_images:
            if _is_image_available(self._config.image):
                _remove_image(self._config.image)

    @property
    def runtime(self) -> RemoteRuntime:
        """Returns the runtime if running.

        Raises:
            DeploymentNotStartedError: If the deployment was not started.
        """
        if self._runtime is None:
            raise DeploymentNotStartedError()
        return self._runtime



import asyncio
from abc import ABC, abstractmethod

from swerex.runtime.abstract import AbstractRuntime, IsAliveResponse

__all__ = ["AbstractDeployment"]


class AbstractDeployment(ABC):
    @abstractmethod
    async def is_alive(self, *, timeout: float | None = None) -> IsAliveResponse:
        """Checks if the runtime is alive. The return value can be
        tested with bool().

        Raises:
            DeploymentNotStartedError: If the deployment was not started.
        """

    @abstractmethod
    async def start(self, *args, **kwargs):
        """Starts the runtime."""

    @abstractmethod
    async def stop(self, *args, **kwargs):
        """Stops the runtime."""

    @property
    @abstractmethod
    def runtime(self) -> AbstractRuntime:
        """Returns the runtime if running.

        Raises:
            DeploymentNotStartedError: If the deployment was not started.
        """

    def __del__(self):
        """Stops the runtime when the object is deleted."""
        print("Stopping runtime because Deployment object is deleted")
        asyncio.run(self.stop())



from abc import ABC, abstractmethod
from typing import Annotated, Literal

from pydantic import BaseModel, Field


class IsAliveResponse(BaseModel):
    """Response to the is_alive request.

    You can test the result with bool().
    """

    is_alive: bool

    message: str = ""
    """Error message if is_alive is False."""

    def __bool__(self) -> bool:
        return self.is_alive


class CreateBashSessionRequest(BaseModel):
    startup_source: list[str] = []
    """Source the following files before running commands.
    The reason this gets a special treatment is that these files
    often overwrite PS1, which we need to reset.
    """
    session: str = "default"
    session_type: Literal["bash"] = "bash"


CreateSessionRequest = Annotated[CreateBashSessionRequest, Field(discriminator="session_type")]
"""Union type for all create session requests. Do not use this directly."""


class CreateBashSessionResponse(BaseModel):
    output: str = ""
    """Output from starting the session."""

    session_type: Literal["bash"] = "bash"


CreateSessionResponse = Annotated[CreateBashSessionResponse, Field(discriminator="session_type")]
"""Union type for all create session responses. Do not use this directly."""


# todo: implement non-output-timeout
class BashAction(BaseModel):
    command: str
    """The command to run."""

    session: str = "default"
    """The session to run the command in."""

    timeout: float | None = None
    """The timeout for the command. None means no timeout."""

    is_interactive_command: bool = False
    """For a non-exiting command to an interactive program
    (e.g., gdb), set this to True."""

    is_interactive_quit: bool = False
    """This will disable checking for exit codes, since the command won't terminate.
    If the command is something like "quit" and should terminate the
    interactive program, set this to False.
    """

    check: bool = False
    """Whether to check for the exit code. If True, we will raise a
    `NonZeroExitCodeError` if the command has a non-zero exit code.
    """

    error_msg: str = ""
    """This error message will be used in the `NonZeroExitCodeError` if the
    command has a non-zero exit code and `check` is True.
    """

    expect: list[str] = []
    """Outputs to expect in addition to the PS1"""

    session_type: Literal["bash"] = "bash"
    """Used for type discrimination. Do not change."""


Action = Annotated[BashAction, Field(discriminator="session_type")]
"""Union type for all actions. Do not use this directly."""


class BashObservation(BaseModel):
    output: str = ""
    exit_code: int | None = None
    failure_reason: str = ""

    expect_string: str = ""
    """Which of the expect strings was matched to terminate the command.
    Empty string if the command timed out etc.
    """

    session_type: Literal["bash"] = "bash"


Observation = Annotated[BashObservation, Field(discriminator="session_type")]
"""Union type for all observations. Do not use this directly."""


class CloseBashSessionRequest(BaseModel):
    session: str = "default"
    session_type: Literal["bash"] = "bash"


CloseSessionRequest = Annotated[CloseBashSessionRequest, Field(discriminator="session_type")]
"""Union type for all close session requests. Do not use this directly."""


class CloseBashSessionResponse(BaseModel):
    session_type: Literal["bash"] = "bash"


CloseSessionResponse = Annotated[CloseBashSessionResponse, Field(discriminator="session_type")]
"""Union type for all close session responses. Do not use this directly."""


class Command(BaseModel):
    """A command to run as a subprocess."""

    command: str | list[str]
    """The command to run. Should be a list of strings (recommended because
    of automatic escaping of spaces etc.) unless you set `shell=True`
    (i.e., exactly like with `subprocess.run()`).
    """

    timeout: float | None = None
    """The timeout for the command. None means no timeout."""

    shell: bool = False
    """Same as the `subprocess.run()` `shell` argument."""

    check: bool = False
    """Whether to check for the exit code. If True, we will raise a
    `CommandFailedError` if the command fails.
    """

    error_msg: str = ""
    """This error message will be used in the `NonZeroExitCodeError` if the
    command has a non-zero exit code and `check` is True.
    """

    env: dict[str, str] | None = None
    """Environment variables to pass to the command."""


class CommandResponse(BaseModel):
    stdout: str = ""
    stderr: str = ""
    exit_code: int | None = None


class ReadFileRequest(BaseModel):
    path: str


class ReadFileResponse(BaseModel):
    content: str = ""


class WriteFileRequest(BaseModel):
    content: str
    path: str


class WriteFileResponse(BaseModel):
    pass


class UploadRequest(BaseModel):
    source_path: str
    target_path: str


class UploadResponse(BaseModel):
    pass


class CloseResponse(BaseModel):
    pass


class _ExceptionTransfer(BaseModel):
    """Helper class to transfer exceptions from the remote runtime to the client."""

    message: str = ""
    class_path: str = ""
    traceback: str = ""


class AbstractRuntime(ABC):
    """This is the main entry point for running stuff.

    It keeps track of all the sessions (individual repls) that are currently open.
    """

    @abstractmethod
    async def is_alive(self, *, timeout: float | None = None) -> IsAliveResponse:
        """Checks if the runtime is alive."""
        pass

    @abstractmethod
    async def create_session(self, request: CreateSessionRequest) -> CreateSessionResponse:
        """Creates a new session."""
        pass

    @abstractmethod
    async def run_in_session(self, action: Action) -> Observation:
        """Runs a command in a session."""
        pass

    @abstractmethod
    async def close_session(self, request: CloseSessionRequest) -> CloseSessionResponse:
        """Closes a shell session."""
        pass

    @abstractmethod
    async def execute(self, command: Command) -> CommandResponse:
        """Executes a command (independent of any shell session)."""
        pass

    @abstractmethod
    async def read_file(self, request: ReadFileRequest) -> ReadFileResponse:
        """Reads a file"""
        pass

    @abstractmethod
    async def write_file(self, request: WriteFileRequest) -> WriteFileResponse:
        """Writes a file"""
        pass

    @abstractmethod
    async def upload(self, request: UploadRequest) -> UploadResponse:
        """Uploads a file"""
        pass

    @abstractmethod
    async def close(self) -> CloseResponse:
        """Closes the runtime."""
        pass



import os

import pytest

from swerex.deployment.fargate import FargateDeployment


@pytest.mark.slow
@pytest.mark.skipif(os.getenv("GITHUB_ACTIONS") == "true", reason="Skipping modal tests in github actions")
async def test_fargate_deployment():
    d = FargateDeployment(
        image="public.ecr.aws/v0o5q4h6/test-namespace/test-repo-1:latest",
        port=8880,
        container_timeout=60 * 15,
    )
    with pytest.raises(RuntimeError):
        await d.is_alive()
    await d.start()
    assert await d.is_alive()
    await d.stop()


@pytest.mark.slow
@pytest.mark.skipif(os.getenv("GITHUB_ACTIONS") == "true", reason="Skipping modal tests in github actions")
async def test_fargate_deployment_ubuntu_base():
    d = FargateDeployment(
        image="ubuntu:24.04",
        port=8880,
        container_timeout=60 * 15,
    )
    with pytest.raises(RuntimeError):
        await d.is_alive()
    await d.start()
    assert await d.is_alive()
    await d.stop()



from typing import Any, Self

from swerex.deployment.abstract import AbstractDeployment
from swerex.deployment.config import RemoteDeploymentConfig
from swerex.exceptions import DeploymentNotStartedError
from swerex.runtime.abstract import IsAliveResponse
from swerex.runtime.config import RemoteRuntimeConfig
from swerex.runtime.remote import RemoteRuntime
from swerex.utils.log import get_logger


class RemoteDeployment(AbstractDeployment):
    def __init__(self, **kwargs: Any):
        """This deployment is only a thin wrapper around the `RemoteRuntime`.
        Use this if you have deployed a runtime somewhere else but want to interact with it
        through the `AbstractDeployment` interface.
        For example, if you have an agent that you usually use with a `DocerkDeployment` interface,
        you sometimes might want to manually start a docker container for debugging purposes.
        Then you can use this deployment to explicitly connect to your manually started runtime.

        Args:
            **kwargs: Keyword arguments (see `RemoteDeploymentConfig` for details).
        """
        self._config = RemoteDeploymentConfig(**kwargs)
        self._runtime: RemoteRuntime | None = None
        self.logger = get_logger("grd")

    @classmethod
    def from_config(cls, config: RemoteDeploymentConfig) -> Self:
        return cls(**config.model_dump())

    @property
    def runtime(self) -> RemoteRuntime:
        """Returns the runtime if running.

        Raises:
            DeploymentNotStartedError: If the deployment was not started.
        """
        if self._runtime is None:
            raise DeploymentNotStartedError()
        return self._runtime

    async def is_alive(self) -> IsAliveResponse:
        """Checks if the runtime is alive. The return value can be
        tested with bool().

        Raises:
            DeploymentNotStartedError: If the deployment was not started.
        """
        return await self.runtime.is_alive()

    async def start(self):
        """Starts the runtime."""
        self.logger.info("Starting remote runtime")
        self._runtime = RemoteRuntime.from_config(
            RemoteRuntimeConfig(
                auth_token=self._config.auth_token,
                host=self._config.host,
                port=self._config.port,
                timeout=self._config.timeout,
            )
        )

    async def stop(self):
        """Stops the runtime."""
        await self.runtime.close()
        self._runtime = None



from typing import Any, Self

from swerex.deployment.abstract import AbstractDeployment
from swerex.deployment.config import DummyDeploymentConfig
from swerex.runtime.abstract import IsAliveResponse
from swerex.runtime.dummy import DummyRuntime


class DummyDeployment(AbstractDeployment):
    def __init__(self, **kwargs: Any):
        """This deployment does nothing.
        Useful for testing.

        Args:
            **kwargs: Keyword arguments (see `DummyDeploymentConfig` for details).
        """
        self._config = DummyDeploymentConfig(**kwargs)
        self._runtime = DummyRuntime()  # type: ignore

    @classmethod
    def from_config(cls, config: DummyDeploymentConfig) -> Self:
        return cls(**config.model_dump())

    async def is_alive(self, *, timeout: float | None = None) -> IsAliveResponse:
        return IsAliveResponse(is_alive=True)

    async def start(self):
        pass

    async def stop(self):
        pass

    @property
    def runtime(self) -> DummyRuntime:
        return self._runtime

    @runtime.setter
    def runtime(self, runtime: DummyRuntime):
        self._runtime = runtime



from pathlib import Path

import pytest

from swerex.runtime.abstract import ReadFileRequest, UploadRequest
from swerex.runtime.local import LocalRuntime


@pytest.fixture
def local_runtime():
    return LocalRuntime()


async def test_upload_file(local_runtime: LocalRuntime, tmp_path: Path):
    file_path = tmp_path / "source.txt"
    file_path.write_text("test")
    tmp_target = tmp_path / "target.txt"
    await local_runtime.upload(UploadRequest(source_path=str(file_path), target_path=str(tmp_target)))
    assert (await local_runtime.read_file(ReadFileRequest(path=str(tmp_target)))).content == "test"


async def test_upload_directory(local_runtime: LocalRuntime, tmp_path: Path):
    dir_path = tmp_path / "source_dir"
    dir_path.mkdir()
    (dir_path / "file1.txt").write_text("test1")
    (dir_path / "file2.txt").write_text("test2")
    tmp_target = tmp_path / "target_dir"
    await local_runtime.upload(UploadRequest(source_path=str(dir_path), target_path=str(tmp_target)))
    assert (await local_runtime.read_file(ReadFileRequest(path=str(tmp_target / "file1.txt")))).content == "test1"
    assert (await local_runtime.read_file(ReadFileRequest(path=str(tmp_target / "file2.txt")))).content == "test2"



import re
import shutil
import subprocess
import time
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Self

import bashlex
import bashlex.ast
import bashlex.errors
import pexpect

from swerex.exceptions import (
    BashIncorrectSyntaxError,
    CommandTimeoutError,
    NoExitCodeError,
    NonZeroExitCodeError,
    SessionDoesNotExistError,
    SessionExistsError,
    SessionNotInitializedError,
)
from swerex.runtime.abstract import (
    AbstractRuntime,
    Action,
    BashAction,
    BashObservation,
    CloseBashSessionResponse,
    CloseResponse,
    CloseSessionRequest,
    CloseSessionResponse,
    Command,
    CommandResponse,
    CreateBashSessionRequest,
    CreateBashSessionResponse,
    CreateSessionRequest,
    CreateSessionResponse,
    IsAliveResponse,
    Observation,
    ReadFileRequest,
    ReadFileResponse,
    UploadRequest,
    UploadResponse,
    WriteFileRequest,
    WriteFileResponse,
)
from swerex.runtime.config import LocalRuntimeConfig
from swerex.utils.log import get_logger

__all__ = ["LocalRuntime", "BashSession"]


def _split_bash_command(inpt: str) -> list[str]:
    r"""Split a bash command with linebreaks, escaped newlines, and heredocs into a list of
    individual commands.

    Args:
        inpt: The input string to split into commands.
    Returns:
        A list of commands as strings.

    Examples:

    "cmd1\ncmd2" are two commands
    "cmd1\\\n asdf" is one command (because the linebreak is escaped)
    "cmd1<<EOF\na\nb\nEOF" is one command (because of the heredoc)
    """
    inpt = inpt.strip()
    if not inpt or all(l.strip().startswith("#") for l in inpt.splitlines()):
        # bashlex can't deal with empty strings or the like :/
        return []
    parsed = bashlex.parse(inpt)
    cmd_strings = []

    def find_range(cmd: bashlex.ast.node) -> tuple[int, int]:
        start = cmd.pos[0]  # type: ignore
        end = cmd.pos[1]  # type: ignore
        for part in getattr(cmd, "parts", []):
            part_start, part_end = find_range(part)
            start = min(start, part_start)
            end = max(end, part_end)
        return start, end

    for cmd in parsed:
        start, end = find_range(cmd)
        cmd_strings.append(inpt[start:end])
    return cmd_strings


def _strip_control_chars(s: str) -> str:
    ansi_escape = re.compile(r"\x1B[@-_][0-?]*[ -/]*[@-~]")
    return ansi_escape.sub("", s)


def _check_bash_command(command: str) -> None:
    """Check if a bash command is valid. Raises BashIncorrectSyntaxError if it's not."""
    _unique_string = "SOUNIQUEEOF"
    cmd = f"/bin/bash -n << '{_unique_string}'\n{command}\n{_unique_string}"
    result = subprocess.run(cmd, shell=True, capture_output=True)
    if result.returncode == 0:
        return
    stdout = result.stdout.decode(errors="backslashreplace")
    stderr = result.stderr.decode(errors="backslashreplace")
    msg = (
        f"Error (exit code {result.returncode}) while checking bash command \n{command!r}\n"
        f"---- Stderr ----\n{stderr}\n---- Stdout ----\n{stdout}"
    )
    raise BashIncorrectSyntaxError(msg)


class Session(ABC):
    @abstractmethod
    async def start(self) -> CreateSessionResponse: ...

    @abstractmethod
    async def run(self, action: Action) -> Observation: ...

    @abstractmethod
    async def close(self) -> CloseSessionResponse: ...


class BashSession(Session):
    _UNIQUE_STRING = "UNIQUESTRING29234"

    def __init__(self, request: CreateBashSessionRequest):
        """This basically represents one REPL that we control.

        It's pretty similar to a `pexpect.REPLWrapper`.
        """
        self.request = request
        self._ps1 = "SHELLPS1PREFIX"
        self._shell: pexpect.spawn | None = None
        self.logger = get_logger(f"RexS ({request.session})")

    @property
    def shell(self) -> pexpect.spawn:
        if self._shell is None:
            msg = "shell not initialized"
            raise RuntimeError(msg)
        return self._shell

    async def start(self) -> CreateBashSessionResponse:
        """Spawn the session, source any startupfiles and set the PS1."""
        self._shell = pexpect.spawn(
            "/bin/bash",
            encoding="utf-8",
            echo=False,
        )
        time.sleep(0.1)
        self.shell.sendline(f"echo '{self._UNIQUE_STRING}'")
        try:
            self.shell.expect(self._UNIQUE_STRING, timeout=1)
        except pexpect.TIMEOUT:
            msg = "timeout while initializing shell"
            raise pexpect.TIMEOUT(msg)
        output = self.shell.before
        cmds = [f"source {path}" for path in self.request.startup_source]
        cmds += [
            f"export PS1='{self._ps1}'",
            "export PS2=''",
            "export PS0=''",
        ]
        cmd = " ; ".join(cmds)
        self.shell.sendline(cmd)
        try:
            self.shell.expect(self._ps1, timeout=1)
        except pexpect.TIMEOUT:
            msg = "timeout while setting PS1"
            raise pexpect.TIMEOUT(msg)
        output += "\n---\n" + _strip_control_chars(self.shell.before)  # type: ignore
        return CreateBashSessionResponse(output=output)

    async def run(self, action: BashAction) -> BashObservation:
        """Run a bash action.

        Raises:
            SessionNotInitializedError: If the shell is not initialized.
            CommandTimeoutError: If the command times out.
            NonZeroExitCodeError: If the command has a non-zero exit code and `action.check` is True.
            NoExitCodeError: If we cannot get the exit code of the command.

        Returns:
            BashObservation: The observation of the command.
        """
        if self.shell is None:
            msg = "shell not initialized"
            raise SessionNotInitializedError(msg)
        if action.is_interactive_command or action.is_interactive_quit:
            return await self._run_interactive(action)
        r = await self._run_normal(action)
        if action.check and r.exit_code != 0:
            msg = (
                f"Command {action.command!r} failed with exit code {r.exit_code}. "
                f"Here is the output:\n{r.output!r}"
            )
            if action.error_msg:
                msg = f"{action.error_msg}: {msg}"
            raise NonZeroExitCodeError(msg)
        return r

    async def _run_interactive(self, action: BashAction) -> BashObservation:
        """Run an interactive action. This is different because we don't seek to
        the PS1 and don't attempt to get the exit code.
        """
        assert self.shell is not None
        self.shell.sendline(action.command)
        expect_strings = action.expect + [self._ps1]
        try:
            expect_index = self.shell.expect(expect_strings, timeout=action.timeout)  # type: ignore
            matched_expect_string = expect_strings[expect_index]
        except pexpect.TIMEOUT as e:
            msg = f"timeout while running command {action.command!r}"
            raise CommandTimeoutError(msg) from e
        output: str = _strip_control_chars(self.shell.before).strip()  # type: ignore
        if action.is_interactive_quit:
            assert not action.is_interactive_command
            self.shell.setecho(False)
            self.shell.waitnoecho()
            self.shell.sendline(f"stty -echo; echo '{self._UNIQUE_STRING}'")
            # Might need two expects for some reason
            self.shell.expect(self._UNIQUE_STRING, timeout=1)
            self.shell.expect(self._ps1, timeout=1)
        else:
            # Interactive command.
            # For some reason, this often times enables echo mode within the shell.
            output = output.lstrip().removeprefix(action.command).strip()

        return BashObservation(output=output, exit_code=0, expect_string=matched_expect_string)

    async def _run_normal(self, action: BashAction) -> BashObservation:
        """Run a normal action. This is the default mode.

        There are three steps to this:

        1. Check if the command is valid
        2. Execute the command
        3. Get the exit code
        """
        assert self.shell is not None
        _check_bash_command(action.command)

        # Part 2: Execute the command

        fallback_terminator = False
        # Running multiple interactive commands by sending them with linebreaks would break things
        # because we get multiple PS1s back to back. Instead we just join them with ;
        # However, sometimes bashlex errors and we can't do this. In this case
        # we add a unique string to the end of the command and then seek to that
        # (which is also somewhat brittle, so we don't do this by default).
        try:
            individual_commands = _split_bash_command(action.command)
        except bashlex.errors.ParsingError as e:
            self.logger.error("Bashlex fail: %s", e)
            action.command += f"\n TMPEXITCODE=$? ; sleep 0.1; echo '{self._UNIQUE_STRING}' ; (exit $TMPEXITCODE)"
            fallback_terminator = True
        else:
            action.command = " ; ".join(individual_commands)
        self.shell.sendline(action.command)
        if not fallback_terminator:
            expect_strings = action.expect + [self._ps1]
        else:
            expect_strings = [self._UNIQUE_STRING]
        try:
            expect_index = self.shell.expect(expect_strings, timeout=action.timeout)  # type: ignore
            matched_expect_string = expect_strings[expect_index]
        except pexpect.TIMEOUT as e:
            msg = f"timeout while running command {action.command!r}"
            raise CommandTimeoutError(msg) from e
        output: str = _strip_control_chars(self.shell.before).strip()  # type: ignore

        # Part 3: Get the exit code

        _exit_code_prefix = "EXITCODESTART"
        _exit_code_suffix = "EXITCODEEND"
        self.shell.sendline(f"\necho {_exit_code_prefix}$?{_exit_code_suffix}")
        try:
            self.shell.expect(_exit_code_suffix, timeout=1)
        except pexpect.TIMEOUT:
            msg = "timeout while getting exit code"
            raise NoExitCodeError(msg)
        exit_code_raw: str = _strip_control_chars(self.shell.before).strip()  # type: ignore
        exit_code = re.findall(f"{_exit_code_prefix}([0-9]+)", exit_code_raw)
        if len(exit_code) != 1:
            msg = f"failed to parse exit code from output {exit_code_raw!r} (command: {action.command!r}, matches: {exit_code})"
            raise NoExitCodeError(msg)
        output += exit_code_raw.split(_exit_code_prefix)[0]
        exit_code = int(exit_code[0])
        # We get at least one more PS1 here.
        try:
            self.shell.expect(self._ps1, timeout=0.1)
        except pexpect.TIMEOUT:
            msg = "Timeout while getting PS1 after exit code extraction"
            raise CommandTimeoutError(msg)
        output = output.replace(self._UNIQUE_STRING, "").replace(self._ps1, "")
        return BashObservation(output=output, exit_code=exit_code, expect_string=matched_expect_string)

    async def close(self) -> CloseSessionResponse:
        if self._shell is None:
            return CloseBashSessionResponse()
        self.shell.close()
        self._shell = None
        return CloseBashSessionResponse()

    def interact(self) -> None:
        """Enter interactive mode."""
        self.shell.interact()


class LocalRuntime(AbstractRuntime):
    def __init__(self, **kwargs: Any):
        """A Runtime that runs locally and actually executes commands in a shell.
        If you are deploying to Modal/Fargate/etc., this class will be running within the docker container
        on Modal/Fargate/etc.

        Args:
            **kwargs: Keyword arguments (see `LocalRuntimeConfig` for details).
        """
        self._config = LocalRuntimeConfig(**kwargs)
        self._sessions: dict[str, Session] = {}

    @classmethod
    def from_config(cls, config: LocalRuntimeConfig) -> Self:
        return cls(**config.model_dump())

    @property
    def sessions(self) -> dict[str, Session]:
        return self._sessions

    async def is_alive(self, *, timeout: float | None = None) -> IsAliveResponse:
        """Checks if the runtime is alive."""
        return IsAliveResponse(is_alive=True)

    async def create_session(self, request: CreateSessionRequest) -> CreateSessionResponse:
        """Creates a new session."""
        if request.session in self.sessions:
            msg = f"session {request.session} already exists"
            raise SessionExistsError(msg)
        if isinstance(request, CreateBashSessionRequest):
            session = BashSession(request)
        else:
            msg = f"unknown session type: {request!r}"
            raise ValueError(msg)
        self.sessions[request.session] = session
        return await session.start()

    async def run_in_session(self, action: Action) -> Observation:
        """Runs a command in a session."""
        if action.session not in self.sessions:
            msg = f"session {action.session!r} does not exist"
            raise SessionDoesNotExistError(msg)
        return await self.sessions[action.session].run(action)

    async def close_session(self, request: CloseSessionRequest) -> CloseSessionResponse:
        """Closes a shell session."""
        if request.session not in self.sessions:
            msg = f"session {request.session!r} does not exist"
            raise SessionDoesNotExistError(msg)
        out = await self.sessions[request.session].close()
        del self.sessions[request.session]
        return out

    async def execute(self, command: Command) -> CommandResponse:
        """Executes a command (independent of any shell session).

        Raises:
            CommandTimeoutError: If the command times out.
            NonZeroExitCodeError: If the command has a non-zero exit code and `check` is True.
        """
        try:
            result = subprocess.run(command.command, shell=command.shell, timeout=command.timeout, env=command.env, capture_output=True)
            r = CommandResponse(
                stdout=result.stdout.decode(errors="backslashreplace"),
                stderr=result.stderr.decode(errors="backslashreplace"),
                exit_code=result.returncode,
            )
        except subprocess.TimeoutExpired as e:
            msg = f"Timeout ({command.timeout}s) exceeded while running command"
            raise CommandTimeoutError(msg) from e
        if command.check and result.returncode != 0:
            msg = (
                f"Command {command.command!r} failed with exit code {result.returncode}. "
                f"Stdout:\n{r.stdout!r}\nStderr:\n{r.stderr!r}"
            )
            if command.error_msg:
                msg = f"{command.error_msg}: {msg}"
            raise NonZeroExitCodeError(msg)
        return r

    async def read_file(self, request: ReadFileRequest) -> ReadFileResponse:
        """Reads a file"""
        content = Path(request.path).read_text()
        return ReadFileResponse(content=content)

    async def write_file(self, request: WriteFileRequest) -> WriteFileResponse:
        """Writes a file"""
        Path(request.path).parent.mkdir(parents=True, exist_ok=True)
        Path(request.path).write_text(request.content)
        return WriteFileResponse()

    async def upload(self, request: UploadRequest) -> UploadResponse:
        """Uploads a file"""
        if Path(request.source_path).is_dir():
            shutil.copytree(request.source_path, request.target_path)
        else:
            shutil.copy(request.source_path, request.target_path)
        return UploadResponse()

    async def close(self) -> CloseResponse:
        """Closes the runtime."""
        for session in self.sessions.values():
            await session.close()
        return CloseResponse()



import pytest

from swerex.deployment import get_deployment
from swerex.deployment.config import (
    DockerDeploymentConfig,
    FargateDeploymentConfig,
    LocalDeploymentConfig,
    ModalDeploymentConfig,
    RemoteDeploymentConfig,
)
from swerex.deployment.docker import DockerDeployment
from swerex.deployment.fargate import FargateDeployment
from swerex.deployment.local import LocalDeployment
from swerex.deployment.modal import ModalDeployment
from swerex.deployment.remote import RemoteDeployment


def test_get_local_deployment():
    deployment = get_deployment(LocalDeploymentConfig())
    assert isinstance(deployment, LocalDeployment)


def test_get_docker_deployment():
    deployment = get_deployment(DockerDeploymentConfig(image="test"))
    assert isinstance(deployment, DockerDeployment)


def test_get_modal_deployment():
    deployment = get_deployment(ModalDeploymentConfig(image="test"))
    assert isinstance(deployment, ModalDeployment)


def test_get_remote_deployment():
    deployment = get_deployment(RemoteDeploymentConfig(auth_token="test"))
    assert isinstance(deployment, RemoteDeployment)


def test_get_fargate_deployment():
    deployment = get_deployment(FargateDeploymentConfig(image="test"))
    assert isinstance(deployment, FargateDeployment)


if __name__ == "__main__":
    pytest.main()



from swerex.deployment.local import LocalDeployment


async def test_local_deployment():
    d = LocalDeployment()
    assert not await d.is_alive()
    await d.start()
    assert await d.is_alive()
    await d.stop()
    assert not await d.is_alive()



import boto3
import json
import hashlib
from urllib.parse import quote


def get_name_hash(prefix: str, obj: dict, max_length: int = 128, hash_length: int = 12) -> str:
    prefix_length = min(max_length, len(prefix))
    if hash_length + prefix_length > max_length:
        raise ValueError(f"Prefix and hash length are too long: {prefix} has length {prefix_length}, hash length is {hash_length}, max length is {max_length}")
    return f"{prefix}-{hashlib.sha256(json.dumps(obj).encode()).hexdigest()[:hash_length]}"


def get_container_name(image_name: str) -> str:
    image_name_sanitized = "".join(c for c in image_name if c.isalnum() or c in "-_")
    if len(image_name_sanitized) <= 255:
        return image_name_sanitized
    return hashlib.sha256(image_name_sanitized.encode()).hexdigest()[:255]


def get_execution_role_arn(execution_role_prefix: str) -> str:
    iam_client = boto3.client('iam')
    
    trust_relationship = {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Effect": "Allow",
                "Principal": {
                    "Service": "ecs-tasks.amazonaws.com"
                },
                "Action": "sts:AssumeRole"
            }
        ]
    }
    
    inline_policy = {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Effect": "Allow",
                "Action": [
                    "logs:CreateLogGroup",
                    "logs:CreateLogStream",
                    "logs:PutLogEvents",
                    "logs:DescribeLogStreams",
                    "secretsmanager:GetSecretValue",
                ],
                "Resource": [
                    "arn:aws:logs:*:*:log-group:/ecs/swe-rex-deployment:*",
                    "arn:aws:logs:*:*:log-group:/ecs/swe-rex-deployment",
                ]
            }
        ]
    }
    role_name = get_name_hash(execution_role_prefix, inline_policy, max_length=64)
    # Check if the role already exists
    role_exists = False
    try:
        role = iam_client.get_role(RoleName=role_name)
        role_exists = True
    except iam_client.exceptions.NoSuchEntityException:
        pass
    
    if not role_exists:
        role = iam_client.create_role(
        RoleName=role_name,
        AssumeRolePolicyDocument=json.dumps(trust_relationship),
        Description="Execution role for ECS tasks",
        MaxSessionDuration=3600,
            Tags=[{'Key': 'origin', 'Value': 'swe-rex-deployment-auto'}],
        )
    # check if policies already attached
    attached_policies = iam_client.list_attached_role_policies(RoleName=role_name)
    attached_policy_arns = [policy['PolicyArn'] for policy in attached_policies['AttachedPolicies']]
    
    if 'arn:aws:iam::aws:policy/service-role/AmazonECSTaskExecutionRolePolicy' not in attached_policy_arns:
        iam_client.attach_role_policy(
        RoleName=role_name,
            PolicyArn='arn:aws:iam::aws:policy/service-role/AmazonECSTaskExecutionRolePolicy'
        )
    if 'arn:aws:iam::aws:policy/AmazonEC2ContainerRegistryReadOnly' not in attached_policy_arns:
        iam_client.attach_role_policy(
            RoleName=role_name,
            PolicyArn='arn:aws:iam::aws:policy/AmazonEC2ContainerRegistryReadOnly'
        )
    
    policy_names = iam_client.list_role_policies(RoleName=role_name)['PolicyNames']
    if 'LogsAndSecretsPolicy' not in policy_names:
        # Add inline policy
        iam_client.put_role_policy(
            RoleName=role_name,
            PolicyName="LogsAndSecretsPolicy",
            PolicyDocument=json.dumps(inline_policy)
        )
    waiter = iam_client.get_waiter('role_exists')
    waiter.wait(RoleName=role_name)
    return role['Role']['Arn']


def get_task_definition(
    image_name: str,
    port: int,
    execution_role_arn: str,
    task_definition_prefix: str,
    log_group: str | None = None,
) -> str:
    ecs_client = boto3.client('ecs')
    task_definition = {
        'executionRoleArn': execution_role_arn,
        'networkMode': 'awsvpc',
        'memory': '2048',
        'cpu': '1 vCPU',
        'containerDefinitions': [
            {
                'name': get_container_name(image_name),
                'image': image_name,
                'portMappings': [
                    {
                        'containerPort': port,
                        'hostPort': port,
                        'protocol': 'tcp'
                    }
                ],
                'essential': True,
                'entryPoint': [
                    '/bin/sh',
                    '-c'
                ],
                'command': [
                    "echo 'hello world'"  # override command with run_task
                ],
            },
        ],
        'requiresCompatibilities': ['FARGATE', 'EC2'],
    }
    if log_group:
        task_definition['containerDefinitions'][0]['logConfiguration'] = {
            'logDriver': 'awslogs',
            'options': {
                'awslogs-group': log_group,
                'awslogs-region': 'us-east-2',
                'awslogs-stream-prefix': 'ecs',
                'awslogs-create-group': 'true',
            }
        }
    family_name = get_name_hash(task_definition_prefix, task_definition, max_length=255)
    # if family exists just return the task definition
    try:
        response = ecs_client.describe_task_definition(taskDefinition=family_name)
        return response['taskDefinition']
    except ecs_client.exceptions.ClientException:
        pass
    
    response = ecs_client.register_task_definition(
        family=family_name,
        **task_definition,
        tags=[{'key': 'origin', 'value': 'swe-rex-deployment-auto'}],
    )
    return response['taskDefinition']


def get_cluster_arn(cluster_name: str) -> str:
    ecs_client = boto3.client('ecs')
    response = ecs_client.create_cluster(
        clusterName=cluster_name,
        tags=[{'key': 'origin', 'value': 'swe-rex-deployment-auto'}],
    )
    return response['cluster']['clusterArn']


def get_default_vpc_and_subnet() -> tuple[str, str]:
    ec2_client = boto3.client('ec2')
    vpcs = ec2_client.describe_vpcs(Filters=[{'Name': 'isDefault', 'Values': ['true']}])
    if not vpcs['Vpcs']:
        raise Exception("No default VPC found")
    vpc_id = vpcs['Vpcs'][0]['VpcId']
    subnets = ec2_client.describe_subnets(Filters=[{'Name': 'vpc-id', 'Values': [vpc_id]}])
    if not subnets['Subnets']:
        raise Exception("No subnets found in the default VPC")
    subnet_id = subnets['Subnets'][0]['SubnetId']
    return vpc_id, subnet_id


def get_security_group(vpc_id: str, port: int, security_group_prefix: str) -> str:
    ec2_client = boto3.client('ec2')
    inbound_rule = {
        'IpProtocol': 'tcp',
        'FromPort': port,
        'ToPort': port,
        'IpRanges': [{'CidrIp': '0.0.0.0/0'}]
    }
    #if it exists, just return the id
    security_group_name = get_name_hash(security_group_prefix, inbound_rule, max_length=255)
    try:
        security_group = ec2_client.describe_security_groups(GroupNames=[security_group_name])
        return security_group['SecurityGroups'][0]['GroupId']
    except ec2_client.exceptions.ClientError:
        pass
    
    security_group = ec2_client.create_security_group(
        GroupName=security_group_name,
        Description='Security group swe rex',
        VpcId=vpc_id,
        TagSpecifications=[
            {
                'ResourceType': 'security-group',
                'Tags': [{'Key': 'origin', 'Value': 'swe-rex-deployment-auto'}],
            },
        ],
    )
    security_group_id = security_group['GroupId']

    # Add an inbound rule to allow traffic on the port
    ec2_client.authorize_security_group_ingress(
        GroupId=security_group_id,
        IpPermissions=[
            inbound_rule
        ]
    )
    return security_group_id


def run_fargate_task(
    command: list[str],
    name: str,
    task_definition_arn: str,
    subnet_id: str,
    security_group_id: str,
    cluster_arn: str,
    vcpus: int = 1,
    memory: int = 2048,
    ephemeral_storage: int = 21,
    **fargate_args,
) -> str:
    run_task_args = {
        'taskDefinition': task_definition_arn,
        'launchType': 'FARGATE',
        'cluster': cluster_arn,
        'networkConfiguration': {
            'awsvpcConfiguration': {
                'subnets': [subnet_id],
                'securityGroups': [security_group_id],
                'assignPublicIp': 'ENABLED'
            }
        },
        'propagateTags': 'TASK_DEFINITION',
    }
    overrides = {
        'containerOverrides': [
            {
                'name': name,
                'command': command,
            },
        ],
        'cpu': f'{vcpus}vCPU',
        'memory': str(memory),
        'ephemeralStorage': {
            'sizeInGiB': ephemeral_storage,
        },
    }
    if fargate_args:
        overrides.update(fargate_args)
        
    if overrides:
        run_task_args['overrides'] = overrides

    ecs_client = boto3.client('ecs')
    response = ecs_client.run_task(
        **run_task_args,
        tags=[{'key': 'origin', 'value': 'swe-rex-deployment-auto'}],
    )
    return response['tasks'][0]['taskArn']
    

def get_public_ip(task_arn: str, cluster_arn: str) -> str:
    ecs_client = boto3.client('ecs')
    task_details = ecs_client.describe_tasks(cluster=cluster_arn, tasks=[task_arn])
    eni_id = task_details['tasks'][0]['attachments'][0]['details'][1]['value']
    ec2_client = boto3.client('ec2')
    eni_details = ec2_client.describe_network_interfaces(NetworkInterfaceIds=[eni_id])
    return eni_details['NetworkInterfaces'][0]['Association']['PublicIp']


def get_cloudwatch_log_url(task_arn: str, task_definition: dict, container_name: str, region: str = 'us-east-2') -> str:
    """
    Get the CloudWatch log URL for a running task.
    
    Args:
        task_arn (str): The ARN of the running task.
        task_definition (dict): The task definition for the running task.
        container_name (str): The name of the container in the task definition.
        
    Returns:
        str: The CloudWatch log URL.
    """
    container_def = task_definition['containerDefinitions'][0]
    log_config = container_def['logConfiguration']['options']
    log_group = log_config['awslogs-group']
    task_id = task_arn.split('/')[-1]
    log_stream = f"{log_config['awslogs-stream-prefix']}/{container_name}/{task_id}"
    
    # Use %25 instead of $25 for encoding
    encoded_log_group = quote(log_group, safe='')
    encoded_log_stream = quote(log_stream, safe='')
    
    return (
        f"https://{region}.console.aws.amazon.com/cloudwatch/home"
        f"?region={region}"
        f"#logsV2:log-groups/log-group/{encoded_log_group}"
        f"/log-events/{encoded_log_stream}"
    )



from typing import Any, Self

from swerex.deployment.abstract import AbstractDeployment
from swerex.deployment.config import LocalDeploymentConfig
from swerex.exceptions import DeploymentNotStartedError
from swerex.runtime.abstract import IsAliveResponse
from swerex.runtime.config import LocalRuntimeConfig
from swerex.runtime.local import LocalRuntime
from swerex.utils.log import get_logger

__all__ = ["LocalDeployment", "LocalDeploymentConfig"]


class LocalDeployment(AbstractDeployment):
    def __init__(
        self,
        **kwargs: Any,
    ):
        """The most boring of the deployment classes.
        This class does nothing but wrap around `Runtime` so you can switch out
        your deployment method.

        Args:
            **kwargs: Keyword arguments (see `LocalDeploymentConfig` for details).
        """
        self._runtime = None
        self.logger = get_logger("deploy")  # type: ignore
        self._config = LocalDeploymentConfig(**kwargs)

    @classmethod
    def from_config(cls, config: LocalDeploymentConfig) -> Self:
        return cls(**config.model_dump())

    async def is_alive(self, *, timeout: float | None = None) -> IsAliveResponse:
        """Checks if the runtime is alive. The return value can be
        tested with bool().

        Raises:
            DeploymentNotStartedError: If the deployment was not started.
        """
        if self._runtime is None:
            return IsAliveResponse(is_alive=False, message="Runtime is None.")
        return await self._runtime.is_alive(timeout=timeout)

    async def start(self):
        """Starts the runtime."""
        self._runtime = LocalRuntime.from_config(LocalRuntimeConfig())

    async def stop(self):
        """Stops the runtime."""
        if self._runtime is not None:
            await self._runtime.close()
            self._runtime = None

    @property
    def runtime(self) -> LocalRuntime:
        """Returns the runtime if running.

        Raises:
            DeploymentNotStartedError: If the deployment was not started.
        """
        if self._runtime is None:
            raise DeploymentNotStartedError()
        return self._runtime



import time
import uuid
from typing import Any, Self

import boto3

from swerex import PACKAGE_NAME, REMOTE_EXECUTABLE_NAME
from swerex.deployment.abstract import AbstractDeployment
from swerex.deployment.config import FargateDeploymentConfig
from swerex.exceptions import DeploymentNotStartedError
from swerex.runtime.abstract import IsAliveResponse
from swerex.runtime.config import RemoteRuntimeConfig
from swerex.runtime.remote import RemoteRuntime
from swerex.utils.aws import (
    get_cloudwatch_log_url,
    get_cluster_arn,
    get_container_name,
    get_default_vpc_and_subnet,
    get_execution_role_arn,
    get_public_ip,
    get_security_group,
    get_task_definition,
    run_fargate_task,
)
from swerex.utils.log import get_logger
from swerex.utils.wait import _wait_until_alive


class FargateDeployment(AbstractDeployment):
    def __init__(
        self,
        **kwargs: Any,
    ):
        self._config = FargateDeploymentConfig(**kwargs)
        self._runtime: RemoteRuntime | None = None
        self._container_process = None
        self._container_name = None
        self.logger = get_logger("deploy")
        # we need to setup ecs and ec2 to run containers
        self._cluster_arn = None
        self._execution_role_arn = None
        self._vpc_id = None
        self._subnet_id = None
        self._task_arn = None
        self._security_group_id = None

    @classmethod
    def from_config(cls, config: FargateDeploymentConfig) -> Self:
        return cls(**config.model_dump())

    def _init_aws(self):
        self._cluster_arn = get_cluster_arn(self._config.cluster_name)
        self._execution_role_arn = get_execution_role_arn(execution_role_prefix=self._config.execution_role_prefix)
        self._task_definition = get_task_definition(
            image_name=self._config.image,
            port=self._config.port,
            execution_role_arn=self._execution_role_arn,
            task_definition_prefix=self._config.task_definition_prefix,
            log_group=self._config.log_group,
        )
        self._vpc_id, self._subnet_id = get_default_vpc_and_subnet()
        self._security_group_id = get_security_group(
            vpc_id=self._vpc_id,
            port=self._config.port,
            security_group_prefix=self._config.security_group_prefix,
        )
        self._container_name = get_container_name(self._config.image)

    def _get_container_name(self) -> str:
        return self._container_name

    @property
    def container_name(self) -> str | None:
        return self._container_name

    async def is_alive(self, *, timeout: float | None = None) -> IsAliveResponse:
        """Checks if the runtime is alive. The return value can be
        tested with bool().

        Raises:
            DeploymentNotStartedError: If the deployment was not started.
        """
        if self._runtime is None or self._task_arn is None:
            raise DeploymentNotStartedError()
        else:
            # check if the task is running
            ecs_client = boto3.client("ecs")
            task_details = ecs_client.describe_tasks(cluster=self._cluster_arn, tasks=[self._task_arn])
            if task_details["tasks"][0]["lastStatus"] != "RUNNING":
                msg = f"Container process not running: {task_details['tasks'][0]['lastStatus']}"
                raise RuntimeError(msg)
        return await self._runtime.is_alive(timeout=timeout)

    async def _wait_until_alive(self, timeout: float):
        return await _wait_until_alive(self.is_alive, timeout=timeout, function_timeout=self._config.container_timeout)

    def _get_command(self, *, token: str) -> list[str]:
        main_command = f"{REMOTE_EXECUTABLE_NAME} --port {self._config.port}"
        fallback_commands = [
            "apt-get update -y",
            "apt-get install pipx -y",
            "pipx ensurepath",
            f"pipx run {PACKAGE_NAME} --port {self._config.port} --auth-token {token}",
        ]
        fallback_script = " && ".join(fallback_commands)
        # Wrap the entire command in bash -c to ensure timeout applies to everything
        inner_command = f"{main_command} || ( {fallback_script} )"
        full_command = f"timeout {self._config.container_timeout}s bash -c '{inner_command}'"
        assert full_command.startswith("timeout "), "command must start with timeout!"
        return [full_command]

    def _get_token(self) -> str:
        return str(uuid.uuid4())

    async def start(
        self,
    ):
        """Starts the runtime."""
        self._init_aws()
        self._container_name = self._get_container_name()
        self.logger.info(f"Starting runtime with container name {self._container_name}")
        token = self._get_token()
        self._task_arn = run_fargate_task(
            command=self._get_command(token=token),
            name=self._container_name,
            task_definition_arn=self._task_definition["taskDefinitionArn"],
            subnet_id=self._subnet_id,
            security_group_id=self._security_group_id,
            cluster_arn=self._cluster_arn,
            **self._config.fargate_args,
        )
        self.logger.info(f"Container task submitted: {self._task_arn} - waiting for it to start...")
        # wait until the container is running
        t0 = time.time()
        ecs_client = boto3.client("ecs")
        waiter = ecs_client.get_waiter("tasks_running")
        waiter.wait(cluster=self._cluster_arn, tasks=[self._task_arn])
        self.logger.info(f"Fargate container started in {time.time() - t0:.2f}s")
        if self._config.log_group:
            try:
                log_url = get_cloudwatch_log_url(
                    task_arn=self._task_arn,
                    task_definition=self._task_definition,
                    container_name=self._container_name,
                )
                self.logger.info(f"Monitor logs at: {log_url}")
            except Exception as e:
                self.logger.warning(f"Failed to get CloudWatch Logs URL: {str(e)}")
        public_ip = get_public_ip(self._task_arn, self._cluster_arn)
        self.logger.info(f"Container public IP: {public_ip}")
        self._runtime = RemoteRuntime.from_config(
            RemoteRuntimeConfig(host=public_ip, port=self._config.port, auth_token=token)
        )
        t0 = time.time()
        await self._wait_until_alive(timeout=self._config.runtime_timeout)
        self.logger.info(f"Runtime started in {time.time() - t0:.2f}s")

    async def stop(self):
        """Stops the runtime."""
        if self._runtime is not None:
            await self._runtime.close()
            self._runtime = None
        if self._task_arn is not None:
            ecs_client = boto3.client("ecs")
            ecs_client.stop_task(task=self._task_arn, cluster=self._cluster_arn)
        self._task_arn = None
        self._container_name = None

    @property
    def runtime(self) -> RemoteRuntime:
        """Returns the runtime if running.

        Raises:
            DeploymentNotStartedError: If the deployment was not started.
        """
        if self._runtime is None:
            msg = "Runtime not started"
            raise RuntimeError(msg)
        return self._runtime



from __future__ import annotations

import logging
import os
from pathlib import PurePath

from rich.logging import RichHandler
from rich.text import Text

_SET_UP_LOGGERS = set()
_ADDITIONAL_HANDLERS = []


def _interpret_level_from_env(level: str | None, *, default=logging.DEBUG) -> int:
    if not level:
        return default
    if level.isnumeric():
        return int(level)
    return getattr(logging, level.upper())


_STREAM_LEVEL = _interpret_level_from_env(os.environ.get("SWE_AGENT_LOG_STREAM_LEVEL"))
_FILE_LEVEL = _interpret_level_from_env(os.environ.get("SWE_AGENT_LOG_FILE_LEVEL"))


class _RichHandlerWithEmoji(RichHandler):
    def __init__(self, emoji: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not emoji.endswith(" "):
            emoji += " "
        self.emoji = emoji

    def get_level_text(self, record: logging.LogRecord) -> Text:
        level_name = record.levelname
        return Text.styled((self.emoji + level_name).ljust(10), f"logging.level.{level_name.lower()}")


def get_logger(name: str, *, emoji: str = "🦖") -> logging.Logger:
    """Get logger. Use this instead of `logging.getLogger` to ensure
    that the logger is set up with the correct handlers.
    """
    logger = logging.getLogger(name)
    if name in _SET_UP_LOGGERS:
        # Already set up
        return logger
    handler = _RichHandlerWithEmoji(
        emoji=emoji,
        show_time=bool(os.environ.get("SWE_AGENT_LOG_TIME", False)),
        show_path=False,
    )
    handler.setLevel(_STREAM_LEVEL)
    logger.setLevel(min(_STREAM_LEVEL, _FILE_LEVEL))
    logger.addHandler(handler)
    logger.propagate = False
    _SET_UP_LOGGERS.add(name)
    for handler in _ADDITIONAL_HANDLERS:
        logger.addHandler(handler)
    return logger


def add_file_handler(path: PurePath | str) -> None:
    """Adds a file handler to all loggers that we have set up
    and all future loggers that will be set up with `get_logger`.
    """
    handler = logging.FileHandler(path)
    formatter = logging.Formatter("%(asctime)s %(levelname)s %(message)s")
    handler.setFormatter(formatter)
    handler.setLevel(_FILE_LEVEL)
    for name in _SET_UP_LOGGERS:
        logger = logging.getLogger(name)
        logger.addHandler(handler)
    _ADDITIONAL_HANDLERS.append(handler)


default_logger = get_logger("swe-agent")



from pathlib import PurePath
from typing import Any, Literal

from pydantic import BaseModel

from swerex.deployment.abstract import AbstractDeployment


class ModalDeploymentConfig(BaseModel):
    image: str | PurePath
    """Image to use for the deployment.
    """
    startup_timeout: float = 0.4
    """The time to wait for the runtime to start."""
    runtime_timeout: float = 1800.0
    """The runtime timeout."""
    modal_sandbox_kwargs: dict[str, Any] = {}
    """Additional arguments to pass to `modal.Sandbox.create`"""


class DockerDeploymentConfig(BaseModel):
    image: str
    """The name of the docker image to use."""
    port: int | None = None
    """The port that the docker container connects to. If None, a free port is found."""
    docker_args: list[str] = []
    """Additional arguments to pass to the docker run command."""
    startup_timeout: float = 60.0
    """The time to wait for the runtime to start."""
    pull: Literal["never", "always", "missing"] = "missing"
    """When to pull docker images."""
    remove_images: bool = False
    """Whether to remove the image after it has stopped."""


class DummyDeploymentConfig(BaseModel):
    pass


class FargateDeploymentConfig(BaseModel):
    image: str
    port: int = 8880
    cluster_name: str = "swe-rex-cluster"
    execution_role_prefix: str = "swe-rex-execution-role"
    task_definition_prefix: str = "swe-rex-task"
    log_group: str | None = "/ecs/swe-rex-deployment"
    security_group_prefix: str = "swe-rex-deployment-sg"
    fargate_args: dict[str, str] = {}
    container_timeout: float = 60 * 15
    runtime_timeout: float = 30


class LocalDeploymentConfig(BaseModel):
    """The port that the runtime connects to."""


class RemoteDeploymentConfig(BaseModel):
    auth_token: str
    """The token to use for authentication."""
    host: str = "http://127.0.0.1"
    """The host to connect to."""
    port: int | None = None
    """The port to connect to."""
    timeout: float = 0.15


DeploymentConfig = (
    LocalDeploymentConfig
    | DockerDeploymentConfig
    | ModalDeploymentConfig
    | FargateDeploymentConfig
    | RemoteDeploymentConfig
    | DummyDeploymentConfig
)


def get_deployment(
    config: DeploymentConfig,
) -> AbstractDeployment:
    # Defer imports to avoid pulling in unnecessary dependencies
    if isinstance(config, DummyDeploymentConfig):
        from swerex.deployment.dummy import DummyDeployment

        return DummyDeployment.from_config(config)
    if isinstance(config, LocalDeploymentConfig):
        from swerex.deployment.local import LocalDeployment

        return LocalDeployment.from_config(config)
    if isinstance(config, DockerDeploymentConfig):
        from swerex.deployment.docker import DockerDeployment

        return DockerDeployment.from_config(config)
    if isinstance(config, ModalDeploymentConfig):
        from swerex.deployment.modal import ModalDeployment

        return ModalDeployment.from_config(config)
    if isinstance(config, FargateDeploymentConfig):
        from swerex.deployment.fargate import FargateDeployment

        return FargateDeployment.from_config(config)
    if isinstance(config, RemoteDeploymentConfig):
        from swerex.deployment.remote import RemoteDeployment

        return RemoteDeployment.from_config(config)

    msg = f"Unknown deployment type: {type(config)}"
    raise ValueError(msg)
